""" NormalizeException simplifies try blocks. """

from pythran.passmanager import Transformation

import gast as ast


class NormalizeException(Transformation):
    '''
    Transform else statement in try except block in nested try except.

    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse("try:print('t')\\nexcept: print('x')\\n\
else: print('e')")
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(NormalizeException, node)
    >>> print(pm.dump(backend.Python, node))
    try:
        print('t')
        try:
            print('e')
        except:
            pass
    except:
        print('x')
    '''
    # FIXME : The transformation is incorrect. Else statement should propagate
    # exception
    def visit_Try(self, node):
        if node.orelse:
            node.body.append(
                ast.Try(
                    node.orelse,
                    [ast.ExceptHandler(None, None, [ast.Pass()])],
                    [], []
                    )
                )
            node.orelse = []
            self.update = True
        if node.finalbody:
            node.body.extend(node.finalbody)
            node.finalbody.append(ast.Raise(None, None))
            self.update = True
            node = ast.Try(
                node.body,
                [ast.ExceptHandler(None, None, node.finalbody)],
                [], [])
            node.finalbody = []
        return node



"""IterTransformation replaces expressions by iterators when possible."""

from pythran.analyses import PotentialIterator, Aliases
from pythran.passmanager import Transformation
from pythran.utils import path_to_attr, path_to_node


EQUIVALENT_ITERATORS = {
    ('builtins', "list"): None,
    ('builtins', "tuple"): None,
    ('numpy', "array"): None,
    ('numpy', "asarray"): None,
    ('numpy', "copy"): None,
}


class IterTransformation(Transformation):

    """
    Replaces expressions by iterators when possible.

    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse('''
    ... def foo(l):
    ...     return builtins.sum(l)
    ... def bar(n):
    ...     return foo(builtins.list(n))
    ... ''')
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(IterTransformation, node)
    >>> print(pm.dump(backend.Python, node))
    def foo(l):
        return builtins.sum(l)
    def bar(n):
        return foo(n)
    """

    def __init__(self):
        """Gather required information."""
        Transformation.__init__(self, PotentialIterator, Aliases)

    def find_matching_builtin(self, node):
        """
        Return matched keyword.

        If the node alias on a correct keyword (and only it), it matches.
        """
        for path in EQUIVALENT_ITERATORS.keys():
            correct_alias = {path_to_node(path)}
            if self.aliases[node.func] == correct_alias:
                return path

    def visit_Call(self, node):
        """Replace function call by its correct iterator if it is possible."""
        if node in self.potential_iterator:
            matched_path = self.find_matching_builtin(node)
            if matched_path is None:
                return self.generic_visit(node)

            # if any kind of specific (~ with more arg) behavior is required
            if len(node.args) != 1:
                return self.generic_visit(node)

            path = EQUIVALENT_ITERATORS[matched_path]
            if path:
                node.func = path_to_attr(path)
            else:
                node = node.args[0]

            self.update = True
        return self.generic_visit(node)



""" ExpandImports replaces imports by their full paths. """

from pythran.passmanager import Transformation
from pythran.utils import path_to_attr, path_to_node
from pythran.conversion import mangle
from pythran.syntax import PythranSyntaxError
from pythran.analyses import Ancestors

import gast as ast


class ExpandImports(Transformation):

    """
    Expands all imports into full paths.

    Attributes
    ----------
    imports : {str}
        Imported module (python base module name)
    symbols : {str : (str,)}
        Matching between used name and real cxx name.

    Examples
    --------
    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse("from math import cos ; cos(2)")
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(ExpandImports, node)
    >>> print(pm.dump(backend.Python, node))
    import math as __pythran_import_math
    __pythran_import_math.cos(2)
    >>> node = ast.parse("from os.path import join ; join('a', 'b')")
    >>> _, node = pm.apply(ExpandImports, node)
    >>> print(pm.dump(backend.Python, node))
    import os as __pythran_import_os
    __pythran_import_os.path.join('a', 'b')
    """

    def __init__(self):
        super(ExpandImports, self).__init__(Ancestors)
        self.imports = set()
        self.symbols = dict()

    def visit_Module(self, node):
        """
        Visit the whole module and add all import at the top level.

        >> import numpy.linalg

        Becomes

        >> import numpy

        """
        node.body = [k for k in (self.visit(n) for n in node.body) if k]
        imports = [ast.Import([ast.alias(i, mangle(i))]) for i in self.imports]
        node.body = imports + node.body
        ast.fix_missing_locations(node)
        return node

    def visit_Import(self, node):
        """ Register imported modules and usage symbols.  """
        for alias in node.names:
            alias_name = tuple(alias.name.split('.'))
            self.imports.add(alias_name[0])
            if alias.asname:
                self.symbols[alias.asname] = alias_name
            else:
                self.symbols[alias_name[0]] = alias_name[:1]
            self.update = True
        return None

    def visit_ImportFrom(self, node):
        """ Register imported modules and usage symbols.  """
        module_path = tuple(node.module.split('.'))
        self.imports.add(module_path[0])
        for alias in node.names:
            path = module_path + (alias.name,)
            self.symbols[alias.asname or alias.name] = path
        self.update = True
        return None

    def visit_FunctionDef(self, node):
        """
        Update import context using overwriting name information.

        Examples
        --------
        >> import foo
        >> import bar
        >> def foo(bar):
        >>     print(bar)

        In this case, neither bar nor foo can be used in the foo function and
        in future function, foo will not be usable.
        """
        self.symbols.pop(node.name, None)
        gsymbols = self.symbols.copy()
        [self.symbols.pop(arg.id, None) for arg in node.args.args]
        self.generic_visit(node)
        self.symbols = gsymbols
        return node

    def visit_Assign(self, node):
        """
        Update import context using overwriting name information.

        Examples
        --------
        >> import foo
        >> def bar():
        >>     foo = 2
        >>     print(foo)

        In this case, foo can't be used after assign.
        """
        if isinstance(node.value, ast.Name) and node.value.id in self.symbols:
            symbol = path_to_node(self.symbols[node.value.id])
            if not getattr(symbol, 'isliteral', lambda: False)():
                for target in node.targets:
                    if not isinstance(target, ast.Name):
                        err = "Unsupported module aliasing"
                        raise PythranSyntaxError(err, target)
                    self.symbols[target.id] = self.symbols[node.value.id]
                return None  # this assignment is no longer needed
        new_node = self.generic_visit(node)
        # no problem if targets contains a subscript, it is not a new assign.
        [self.symbols.pop(t.id, None)
         for t in new_node.targets if isinstance(t, ast.Name)]
        return new_node

    def visit_Name(self, node):
        """
        Replace name with full expanded name.

        Examples
        --------
        >> from numpy.linalg import det

        >> det(a)

        Becomes

        >> numpy.linalg.det(a)
        """
        self.generic_visit(node)
        if node.id in self.symbols:
            symbol = path_to_node(self.symbols[node.id])
            if not getattr(symbol, 'isliteral', lambda: False)():
                parent = self.ancestors[node][-1]
                blacklist = (ast.Tuple,
                             ast.List,
                             ast.Set,
                             ast.Return)
                if isinstance(parent, blacklist):
                    raise PythranSyntaxError(
                        "Unsupported module identifier manipulation",
                        node)
            new_node = path_to_attr(self.symbols[node.id])
            new_node.ctx = node.ctx
            ast.copy_location(new_node, node)
            return new_node
        return node



"""
Ancestors computes the ancestors of each node
"""

from pythran.passmanager import ModuleAnalysis

class Ancestors(ModuleAnalysis):
    '''
    Associate each node with the list of its ancestors

    Based on the tree view of the AST: each node has the Module as parent.
    The result of this analysis is a dictionary with nodes as key,
    and list of nodes as values.
    '''

    def __init__(self):
        self.result = dict()
        self.current = tuple()
        super(Ancestors, self).__init__()

    def generic_visit(self, node):
        self.result[node] = current = self.current
        self.current += node,
        super(Ancestors, self).generic_visit(node)
        self.current = current

    visit = generic_visit


class AncestorsWithBody(Ancestors):

    visit = ModuleAnalysis.visit

    def visit_metadata(self, node):
        if hasattr(node, 'metadata'):
            self.generic_visit(node.metadata)

    def visit_body(self, body):
        body_as_tuple = tuple(body)
        self.result[body_as_tuple] = current = self.current
        self.current += body_as_tuple,
        for stmt in body:
            self.generic_visit(stmt)
        self.current = current

    def visit_If(self, node):
        self.result[node] = current = self.current
        self.current += node,
        self.generic_visit(node.test)
        self.visit_metadata(node)
        self.visit_body(node.body)
        self.visit_body(node.orelse)
        self.current = current

    def visit_While(self, node):
        self.result[node] = current = self.current
        self.current += node,
        self.generic_visit(node.test)
        self.visit_metadata(node)
        self.visit_body(node.body)
        self.visit_body(node.orelse)
        self.current = current

    def visit_For(self, node):
        self.result[node] = current = self.current
        self.current += node,
        self.generic_visit(node.target)
        self.generic_visit(node.iter)
        self.visit_metadata(node)
        self.visit_body(node.body)
        self.visit_body(node.orelse)
        self.current = current

    def visit_Try(self, node):
        self.result[node] = current = self.current
        self.current += node,
        self.visit_metadata(node)
        self.visit_body(node.body)
        for handler in node.handlers:
            self.generic_visit(handler)
        self.visit_body(node.orelse)
        self.visit_body(node.finalbody)
        self.current = current



'''
This modules contains OpenMP-related stuff.
    * OMPDirective is used to represent OpenMP annotations in the AST
    * GatherOMPData turns OpenMP-like string annotations into metadata
'''

from pythran.passmanager import Transformation
import pythran.metadata as metadata
from pythran.spec import parse_pytypes
from pythran.types.conversion import pytype_to_ctype
from pythran.utils import isstr

from gast import AST
import gast as ast
import re


keywords = {
    'atomic',
    'barrier',
    'capture',
    'cancel',
    'collapse',
    'copyin',
    'copyprivate',
    'critical',
    'declare',
    'default',
    'final',
    'firstprivate',
    'flush',
    'for',
    'if',
    'initializer',
    'lastprivate',
    'master',
    'mergeable',
    'none',
    'nowait',
    'num_threads',
    'omp',
    'ordered',
    'parallel',
    'private',
    'read',
    'reduction',
    'schedule',
    'section',
    'sections',
    'shared',
    'simd',
    'single',
    'task',
    'taskwait',
    'taskyield',
    'threadprivate',
    'untied',
    'update',
    'write'
}

declare_keywords = {
    'omp_in',
    'omp_init',
    'omp_orig',
    'omp_out',
    'omp_priv',
}

reserved_contex = {
    'critical',
    'declare',
    'default',
    'schedule',
    'reduction',
}


def is_declare_typename(s, offset, bounds):
    start = s.rfind(':', 0, offset - 1)
    stop = s.rfind(':', offset + 1)
    if start > 0 and stop > 0:
        bounds.extend((start + 1, stop))
        return True
    else:
        return False


class OMPDirective(AST):
    '''Turn a string into a context-dependent metadata.
    >>> o = OMPDirective("omp for private(a,b) shared(c)")
    >>> o.s
    'omp for private({},{}) shared({})'
    >>> [ type(dep) for dep in o.deps ]
    [<class 'gast.gast.Name'>, <class 'gast.gast.Name'>, \
<class 'gast.gast.Name'>]
    >>> [ dep.id for dep in o.deps ]
    ['a', 'b', 'c']
    '''

    def __init__(self, *args):  # no positional argument to be deep copyable
        super(OMPDirective, self).__init__()
        if not args:
            return

        self.deps = []
        self.private_deps = []
        self.shared_deps = []


        def tokenize(s):
            '''A simple contextual "parser" for an OpenMP string'''
            # not completely satisfying if there are strings in if expressions
            out = ''
            par_count = 0
            curr_index = 0
            in_reserved_context = False
            in_declare = False
            in_shared = in_private = False
            while curr_index < len(s):
                bounds = []
                if in_declare and is_declare_typename(s, curr_index, bounds):
                    start, stop = bounds
                    pytypes = parse_pytypes(s[start:stop])
                    out += ', '.join(map(pytype_to_ctype, pytypes))
                    curr_index = stop
                    continue
                m = re.match(r'^([a-zA-Z_]\w*)', s[curr_index:])
                if m:
                    word = m.group(0)
                    curr_index += len(word)
                    if(in_reserved_context or
                         (in_declare and word in declare_keywords) or
                         (par_count == 0 and word in keywords)):
                        out += word
                        in_reserved_context = word in reserved_contex
                        in_declare |= word == 'declare'
                        in_private |= word == 'private'
                        in_shared |= word == 'shared'
                    else:
                        out += '{}'
                        self.deps.append(ast.Name(word, ast.Load(),
                                                  None, None))
                        isattr = re.match(r'^\s*(\.\s*[a-zA-Z_]\w*)', s[curr_index:])
                        if isattr:
                            attr = isattr.group(0)
                            curr_index += len(attr)
                            self.deps[-1] = ast.Attribute(self.deps[-1],
                                                          attr[1:], ast.Load())
                        if in_private:
                            self.private_deps.append(self.deps[-1])
                        if in_shared:
                            self.shared_deps.append(self.deps[-1])
                elif s[curr_index] == '(':
                    par_count += 1
                    curr_index += 1
                    out += '('
                elif s[curr_index] == ')':
                    par_count -= 1
                    curr_index += 1
                    out += ')'
                    if par_count == 0:
                        in_reserved_context = False
                        in_shared = in_private = False
                else:
                    if s[curr_index] in ',:':
                        in_reserved_context = False
                    out += s[curr_index]
                    curr_index += 1
            return out

        self.s = tokenize(args[0])
        self._fields = ('deps', 'shared_deps', 'private_deps')


##
class GatherOMPData(Transformation):
    '''Walks node and collect string comments looking for OpenMP directives.'''

    # there is a special handling for If and Expr, so not listed here
    statements = ("FunctionDef", "Return", "Delete", "Assign", "AugAssign",
                  "Print", "For", "While", "Raise", "TryExcept", "TryFinally",
                  "Assert", "Import", "ImportFrom", "Pass", "Break",)

    # these fields hold statement lists
    statement_lists = ("body", "orelse", "finalbody",)

    def __init__(self):
        Transformation.__init__(self)
        # Remap self.visit_XXXX() to self.attach_data() generic method
        for s in GatherOMPData.statements:
            setattr(self, "visit_" + s, self.attach_data)
        self.current = list()

    def isompdirective(self, node):
        return isstr(node) and node.value.startswith("omp ")

    def visit_Expr(self, node):
        if self.isompdirective(node.value):
            self.current.append(node.value.value)
            return None
        else:
            self.attach_data(node)
            return node

    def visit_If(self, node):
        if self.isompdirective(node.test):
            self.visit(ast.Expr(node.test))
            return self.visit(ast.If(ast.Constant(1, None),
                                     node.body, node.orelse))
        else:
            return self.attach_data(node)

    def attach_data(self, node):
        '''Generic method called for visit_XXXX() with XXXX in
        GatherOMPData.statements list

        '''
        if self.current:
            for curr in self.current:
                md = OMPDirective(curr)
                metadata.add(node, md)
            self.current = list()
        # add a Pass to hold some directives
        for field_name, field in ast.iter_fields(node):
            if field_name in GatherOMPData.statement_lists:
                if(field and
                   isinstance(field[-1], ast.Expr) and
                   self.isompdirective(field[-1].value)):
                    field.append(ast.Pass())
        self.generic_visit(node)

        # add an If to hold scoping OpenMP directives
        directives = metadata.get(node, OMPDirective)
        field_names = {n for n, _ in ast.iter_fields(node)}
        has_no_scope = field_names.isdisjoint(GatherOMPData.statement_lists)
        if directives and has_no_scope:
            # some directives create a scope, but the holding stmt may not
            # artificially create one here if needed
            sdirective = ''.join(d.s for d in directives)
            scoping = ('parallel', 'task', 'section')
            if any(s in sdirective for s in scoping):
                metadata.clear(node, OMPDirective)
                node = ast.If(ast.Constant(1, None), [node], [])
                for directive in directives:
                    metadata.add(node, directive)

        return node



"""
ExtendedSyntaxCheck performs various syntax checks on the pythran AST.
"""

from pythran.passmanager import ModuleAnalysis
from pythran.analyses import StrictAliases, ArgumentEffects
from pythran.syntax import PythranSyntaxError
from pythran.intrinsic import ConstantIntr, FunctionIntr
from pythran import metadata

import gast as ast


def is_global_constant(node):
    if isinstance(node, ConstantIntr):
        return True

    if not isinstance(node, ast.FunctionDef):
        return False

    return metadata.get(node.body[0], metadata.StaticReturn)


def is_global(node):
    return (isinstance(node, (FunctionIntr, ast.FunctionDef)) or
            is_global_constant(node))


class ExtendedSyntaxCheck(ModuleAnalysis):
    """
    Perform advanced syntax checking, based on strict aliases analysis:
    - is there a function redefinition?
    - is there a function call that does not match the called expression arity?
    - is there an operation that updates a global variable?
    """

    def __init__(self):
        self.result = None
        self.update = False
        self.inassert = False
        self.functions = set()
        ModuleAnalysis.__init__(self, StrictAliases, ArgumentEffects)

    def check_global_with_side_effect(self, node, arg):
        if not isinstance(arg, ast.Call):
            return
        try:
            aliases = self.strict_aliases[arg.func]
        except KeyError:
            return

        for alias in aliases:
            if is_global_constant(alias):
                raise PythranSyntaxError(
                    ("Cannot modify '{}': global variables are constant "
                     "in pythran.").format(alias.name),
                    arg.func)

    def visit_FunctionDef(self, node):
        if node.name in self.functions:
            raise PythranSyntaxError("Function {} redefined".format(
                node.name),
                node)
        else:
            self.functions.add(node.name)
        self.generic_visit(node)

    def check_assert_with_side_effect(self, node, arg):
        if self.inassert:
            raise PythranSyntaxError("Cannot call a function with side effect "
                                     "in an assert", node)

    def visit_Assert(self, node):
        self.inassert = True
        self.generic_visit(node)
        self.inassert = False

    def is_immutable_constant(self, node):
        if isinstance(node, ast.Constant):
            return True

        if isinstance(node, ast.Tuple):
            return all(self.is_immutable_constant(elt) for elt in node.elts)

        if isinstance(node, ast.UnaryOp):
            return self.is_immutable_constant(node.operand)

        if isinstance(node, ast.Call):
            target = getattr(node, 'func', node)
            try:
                aliases = self.strict_aliases[target]
            except KeyError:
                return False

            if not aliases:
                return False

            if all(is_global_constant(alias) for alias in aliases):
                return True

        if isinstance(node, ast.Attribute):
            target = getattr(node, 'func', node)
            try:
                aliases = self.strict_aliases[target]
            except KeyError:
                return False

            if not aliases:
                return False

            if all(is_global(alias) for alias in aliases):
                return True

        if isinstance(node, ast.Name):
            try:
                aliases = self.strict_aliases[node]
            except KeyError:
                return False

            if all(isinstance(alias, ast.FunctionDef) for alias in aliases):
                return True

        return False

    def visit_arguments(self, node):
        self.generic_visit(node)
        for arg_default in node.defaults:
            if not self.is_immutable_constant(arg_default):
                raise PythranSyntaxError(
                    "Pythran does not support mutable default values. Use a "
                    "`None' default and set the value at runtime instead.",
                    arg_default)

    def visit_Call(self, node):
        self.generic_visit(node)
        func = node.func
        try:
            aliases = self.strict_aliases[func]
        except KeyError:
            raise PythranSyntaxError(
                "Call to unknown function `{}`, it's a trap!"
                .format(getattr(func, 'id', None) or func),
                node)

        argument_effects = set()

        for alias in aliases:
            # look for effect on arguments to prepare check on globals
            try:
                func_aes = self.argument_effects[alias]
                for i, effect in enumerate(func_aes):
                    if effect:
                        argument_effects.add(i)
            except KeyError:
                pass

            if not isinstance(alias, ast.FunctionDef):
                continue
            ubound = len(alias.args.args)
            lbound = ubound - len(alias.args.defaults)
            call_args_count = len(node.args) + len(node.keywords)
            if lbound <= call_args_count <= ubound:
                continue

            if lbound == ubound:
                msg = 'Invalid call to {}: expected {} arguments, got {}'
                msg = msg.format(alias.name,
                                 len(alias.args.args),
                                 len(node.args)
                                 )
            else:
                msg = ('Invalid {} call: '
                       'expected between {} and {} arguments, got {}')
                msg = msg.format(alias.name,
                                 lbound, ubound,
                                 len(node.args)
                                 )
            raise PythranSyntaxError(msg, node)

        # check for effects on globals
        for i, arg in enumerate(node.args):
            if i not in argument_effects:
                continue
            self.check_global_with_side_effect(node, arg)
            self.check_assert_with_side_effect(node, arg)



""" NormalizeIsNone detects is None patterns. """

from pythran.passmanager import Transformation
from pythran.analyses import Ancestors
from pythran.syntax import PythranSyntaxError
from functools import reduce

import gast as ast


def is_none(expr):
    # py3
    if isinstance(expr, ast.Constant) and expr.value is None:
        return True

    # py2
    if not isinstance(expr, ast.Attribute):
        return False

    return expr.attr == "None"


def is_is_none(expr):
    if not isinstance(expr, ast.Compare):
        return None

    if len(expr.ops) != 1:
        exprs = [expr.left] + expr.comparators
        if any(is_none(expr) for expr in exprs):
            raise PythranSyntaxError("is None in complex condition", expr)
        return None

    if not isinstance(expr.ops[0], (ast.Eq, ast.Is)):
        return None

    if is_none(expr.left):
        return expr.comparators[0]

    if is_none(expr.comparators[0]):
        return expr.left

    return None


def is_is_not_none(expr):
    if not isinstance(expr, ast.Compare):
        return None

    if len(expr.ops) != 1:
        exprs = [expr.left] + expr.comparators
        if any(is_none(expr) for expr in exprs):
            raise PythranSyntaxError("is None in complex condition", expr)
        return None

    if not isinstance(expr.ops[0], (ast.NotEq, ast.IsNot)):
        return None

    if is_none(expr.left):
        return expr.comparators[0]

    if is_none(expr.comparators[0]):
        return expr.left

    return None


class NormalizeIsNone(Transformation):

    table = {ast.And: ast.BitAnd, ast.Or: ast.BitOr}

    def __init__(self):
        super(NormalizeIsNone, self).__init__(Ancestors)

    @staticmethod
    def match_is_none(node):
        noned_var = is_is_none(node)
        if noned_var is None:
            noned_var = is_is_not_none(node)
            negated = noned_var is not None
        else:
            negated = False
        return noned_var, negated

    def visit_BoolOp(self, node):
        values = list(node.values)
        self.generic_visit(node)
        if any(x != y for x, y in zip(values, node.values)):
            self.update = True
            expr = reduce(lambda x, y:
                          ast.BinOp(x,
                                    NormalizeIsNone.table[type(node.op)](), y),
                          node.values)
            return expr
        else:
            return node

    def visit_Compare(self, node):
        self.generic_visit(node)
        noned_var, negated = self.match_is_none(node)
        if noned_var is None:
            return node
        call = ast.Call(
            ast.Attribute(
                ast.Attribute(
                    ast.Name('builtins', ast.Load(), None, None),
                    'pythran',
                    ast.Load()
                ),
                'is_none',
                ast.Load()),
            [noned_var], [])

        self.update = True

        if negated:
            return ast.UnaryOp(ast.Not(), call)
        else:
            return call



"""
NodeCount counts the number of nodes in a node
"""

from pythran.passmanager import NodeAnalysis


class NodeCount(NodeAnalysis):
    """
    Count the number of nodes included in a node

    This has nothing to do with execution time or whatever,
    its mainly use is to prevent the AST from growing too much when unrolling

    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse("if 1: return 3")
    >>> pm = passmanager.PassManager("test")
    >>> print(pm.gather(NodeCount, node))
    5
    """

    def __init__(self):
        self.result = 0
        super(NodeCount, self).__init__()

    def generic_visit(self, node):
        self.result += 1
        super(NodeCount, self).generic_visit(node)



""" FalsePolymorphism try to rename variable to avoid false polymorphism."""

from pythran.passmanager import Transformation
from pythran.analyses import DefUseChains, UseDefChains, Identifiers

import gast as ast


class FalsePolymorphism(Transformation):

    """
    Rename variable when possible to avoid false polymorphism.

    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse("def foo(): a = 12; a = 'babar'")
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(FalsePolymorphism, node)
    >>> print(pm.dump(backend.Python, node))
    def foo():
        a = 12
        a_ = 'babar'
    """

    def __init__(self):
        super(FalsePolymorphism, self).__init__(DefUseChains, UseDefChains)

    def visit_FunctionDef(self, node):

        # reset available identifier names
        # removing local identifiers from the list so that first occurrence can
        # actually use the slot
        identifiers = self.gather(Identifiers, node)
        for def_ in self.def_use_chains.locals[node]:
            try:
                identifiers.remove(def_.name())
            except KeyError:
                pass

        # compute all reachable nodes from each def. This builds a bag of def
        # that should have the same name
        visited_defs = set()
        for def_ in self.def_use_chains.locals[node]:
            if def_ in visited_defs:
                continue

            associated_defs = set()

            # fill the bag of associated defs, going through users and defs
            to_process = [def_]
            while to_process:
                curr = to_process.pop()
                if curr in associated_defs:
                    continue
                if curr.name() != def_.name():
                    continue
                associated_defs.add(curr)
                for u in curr.users():
                    to_process.append(u)
                curr_udc = (d for d in self.use_def_chains.get(curr.node, [])
                            if isinstance(d.node, ast.Name))
                to_process.extend(curr_udc)

            visited_defs.update(associated_defs)

            # find a new identifier
            local_identifier = def_.name()
            name = local_identifier
            while name in identifiers:
                name += "_"
            identifiers.add(name)

            # don't rename first candidate
            if name == local_identifier:
                continue

            # actual renaming of each node in the bag
            self.update = True
            for d in associated_defs:
                dn = d.node
                if isinstance(dn, ast.Name) and dn.id == local_identifier:
                    dn.id = name
        return node




""" Comprehension patterns transforms list comprehension into intrinsics.  """

from pythran.analyses import OptimizableComprehension
from pythran.passmanager import Transformation
from pythran.transformations.normalize_tuples import ConvertToTuple
from pythran.conversion import mangle
from pythran.utils import attr_to_path, path_to_attr

import gast as ast


class ComprehensionPatterns(Transformation):
    '''
    Transforms list comprehension into intrinsics.
    >>> import gast as ast
    >>> from pythran import passmanager, backend
    >>> node = ast.parse("def foo(y) : return (x for x in y)")
    >>> pm = passmanager.PassManager("test")
    >>> _, node = pm.apply(ComprehensionPatterns, node)
    >>> 'map' in pm.dump(backend.Python, node)
    True

    >>> node = ast.parse("def foo(y) : return [0 for _ in builtins.range(y)]")
    >>> _, node = pm.apply(ComprehensionPatterns, node)
    >>> print(pm.dump(backend.Python, node))
    def foo(y):
        return ([0] * builtins.len(builtins.range(y)))
    '''

    def __init__(self):
        Transformation.__init__(self, OptimizableComprehension)

    def visit_Module(self, node):
        self.use_itertools = False
        self.generic_visit(node)
        if self.use_itertools:
            import_alias = ast.alias(name='itertools',
                                     asname=mangle('itertools'))
            importIt = ast.Import(names=[import_alias])
            node.body.insert(0, importIt)
        return node

    def make_Iterator(self, gen):
        if gen.ifs:
            ldFilter = ast.Lambda(
                ast.arguments([ast.Name(gen.target.id, ast.Param(),
                                        None, None)],
                              [], None, [], [], None, []),
                ast.BoolOp(ast.And(), gen.ifs)
                if len(gen.ifs) > 1 else gen.ifs[0])
            ifilterName = ast.Attribute(
                value=ast.Name(id='builtins',
                               ctx=ast.Load(),
                               annotation=None, type_comment=None),
                attr='filter', ctx=ast.Load())
            return ast.Call(ifilterName, [ldFilter, gen.iter], [])
        else:
            return gen.iter

    def visitComp(self, node, make_attr):

        if node in self.optimizable_comprehension:
            self.update = True
            self.generic_visit(node)

            iters = [self.make_Iterator(gen) for gen in node.generators]
            variables = [ast.Name(gen.target.id, ast.Param(), None, None)
                         for gen in node.generators]

            # If dim = 1, product is useless
            if len(iters) == 1:
                iterAST = iters[0]
                varAST = ast.arguments([variables[0]], [],
                                       None, [], [], None, [])
            else:
                self.use_itertools = True
                prodName = ast.Attribute(
                    value=ast.Name(id=mangle('itertools'),
                                   ctx=ast.Load(),
                                   annotation=None, type_comment=None),
                    attr='product', ctx=ast.Load())

                varid = variables[0].id  # retarget this id, it's free
                renamings = {v.id: (i,) for i, v in enumerate(variables)}
                node.elt = ConvertToTuple(varid, renamings).visit(node.elt)
                iterAST = ast.Call(prodName, iters, [])
                varAST = ast.arguments([ast.Name(varid, ast.Param(),
                                                 None, None)],
                                       [], None, [], [], None, [])

            ldBodymap = node.elt
            ldmap = ast.Lambda(varAST, ldBodymap)

            return make_attr(ldmap, iterAST)

        else:
            return self.generic_visit(node)

    def visit_ListComp(self, node):
        def makeattr(*args):
            r = ast.Attribute(
                value=ast.Name(id='builtins',
                               ctx=ast.Load(),
                               annotation=None,
                               type_comment=None),
                attr='map', ctx=ast.Load())
            r = ast.Call(r, list(args), [])
            r = ast.Call(ast.Attribute(ast.Name('builtins', ast.Load(),
                                                None, None),
                                       'list', ast.Load()),
                         [r], [])
            return r

        if isinstance(node.elt, ast.Constant) and len(node.generators) == 1:
            gen = node.generators[0]
            if not gen.ifs and isinstance(gen.iter, ast.Call):
                try:
                    path = attr_to_path(gen.iter.func)[1]
                    range_path = 'pythonic', 'builtins', 'functor', 'range'
                    if path == range_path and len(gen.iter.args) == 1:
                        self.update = True
                        return ast.BinOp(
                            ast.List([node.elt], ast.Load()),
                            ast.Mult(),
                            ast.Call(path_to_attr(('builtins', 'len')),
                                     [gen.iter],
                                     []))
                except TypeError:
                    pass

        return self.visitComp(node, makeattr)

    def visit_GeneratorExp(self, node):
        def makeattr(*args):
            return ast.Call(ast.Attribute(
                value=ast.Name(id='builtins',
                               ctx=ast.Load(),
                               annotation=None, type_comment=None),
                attr='map', ctx=ast.Load()), list(args), [])
        return self.visitComp(node, makeattr)



"""
Whether a list usage makes it a candidate for fixed-size-list

This could be a type information, but it seems easier to implement it that way
"""
from pythran.passmanager import FunctionAnalysis
from pythran.tables import MODULES

import gast as ast


class FixedSizeList(FunctionAnalysis):

    def __init__(self):
        self.result = set()
        from pythran.analyses import Aliases, DefUseChains, Ancestors
        from pythran.analyses import ArgumentEffects
        super(FixedSizeList, self).__init__(Aliases, DefUseChains, Ancestors,
                                            ArgumentEffects)

    def is_fixed_size_list_def(self, node):
        if isinstance(node, ast.List):
            return True

        if not isinstance(node, ast.Call):
            return False

        return all(alias == MODULES['builtins']['list']
                   for alias in self.aliases[node.func])

    def is_safe_call(self, node, index):
        func_aliases = list(self.aliases[node])
        for alias in func_aliases:
            if isinstance(alias, ast.Call):
                if not self.is_safe_call(alias.args[0],
                                         index + len(alias.args) - 1):
                    return False

            if alias in self.argument_effects:
                func_aes = self.argument_effects[alias]
                if func_aes[index]:
                    return False
        return True

    def is_safe_use(self, use):
        parent = self.ancestors[use.node][-1]

        OK = ast.Subscript, ast.BinOp
        if isinstance(parent, OK):
            return True

        if isinstance(parent, ast.Call):
            n = parent.args.index(use.node)
            return self.is_safe_call(parent.func, n)

        return False

    def visit_Assign(self, node):
        self.generic_visit(node)
        if not self.is_fixed_size_list_def(node.value):
            return

        for target in node.targets:
            def_ = self.def_use_chains.chains[target]
            if any(not self.is_safe_use(u) for u in def_.users()):
                break
            if not isinstance(target, ast.Name):
                continue
            if sum(1 for d in self.def_use_chains.locals[self.ctx.function]
                   if d.name() == target.id) > 1:
                break
        else:
            self.result.add(node.value)

    def visit_Call(self, node):
        self.generic_visit(node)
        for i, arg in enumerate(node.args):
            if not self.is_fixed_size_list_def(arg):
                continue
            if self.is_safe_call(node.func, i):
                self.result.add(arg)



"""
This submodule contains all the transformations passes offered in Pythran.

This file is just for convenience and turns the import from

import transformations.xxxxx.xxxxx

into

import transformations.xxxxx
"""


from .expand_builtins import ExpandBuiltins
from .expand_globals import ExpandGlobals
from .expand_import_all import ExpandImportAll
from .expand_imports import ExpandImports
from .extract_doc_strings import ExtractDocStrings
from .false_polymorphism import FalsePolymorphism
from .handle_import import HandleImport
from .normalize_compare import NormalizeCompare
from .normalize_exception import NormalizeException
from .normalize_ifelse import NormalizeIfElse
from .normalize_is_none import NormalizeIsNone
from .normalize_method_calls import NormalizeMethodCalls
from .normalize_return import NormalizeReturn
from .normalize_static_if import NormalizeStaticIf, SplitStaticExpression
from .normalize_tuples import NormalizeTuples
from .normalize_typeis import NormalizeTypeIs
from .remove_comprehension import RemoveComprehension
from .remove_lambdas import RemoveLambdas
from .remove_nested_functions import RemoveNestedFunctions
from .unshadow_parameters import UnshadowParameters
from .remove_named_arguments import RemoveNamedArguments
from .remove_fstrings import RemoveFStrings



""" GlobalDeclarations gathers top-level declarations. """

import gast as ast
from pythran.passmanager import ModuleAnalysis

class GlobalDeclarations(ModuleAnalysis):

    """ Gather all kind of identifier defined at global scope.

    >>> import gast as ast
    >>> from pythran import passmanager
    >>> from pythran.analyses import GlobalDeclarations
    >>> node = ast.parse('''
    ... import math
    ... import math as maths
    ... from math import cos
    ... c = 12
    ... def foo(a):
    ...     b = a + 1''')
    >>> pm = passmanager.PassManager("test")
    >>> sorted(pm.gather(GlobalDeclarations, node).keys())
    ['c', 'cos', 'foo', 'math', 'maths']

    """

    def __init__(self):
        """ Result is an identifier with matching definition. """
        self.result = dict()
        super(GlobalDeclarations, self).__init__()

    def visit_FunctionDef(self, node):
        """ Import module define a new variable name. """
        self.result[node.name] = node

    def visit_Import(self, node):
        for alias in node.names:
            self.result[alias.asname or alias.name] = alias

    def visit_ImportFrom(self, node):
        for alias in node.names:
            self.result[alias.asname or alias.name] = alias

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.result[node.id] = node
